/* cexception.c - Raising C errno based exceptions

   Copyright (c) 2010-2011 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

#include <errno.h>
#include "alore.h"


typedef struct {
    int errNum;
    const char *message;
} ErrorEntry;


#define NUM_ERROR_MESSAGES (sizeof(ErrorMessages) / sizeof(ErrorEntry))


static ErrorEntry ErrorMessages[] = {
#ifdef E2BIG
    { E2BIG, "Argument list too long" }, /* FIX: overflow?? */
#endif
#ifdef EACCES
    { EACCES, "Permission denied" },
#endif
#ifdef EADDRINUSE
    { EADDRINUSE, "Address in use" },
#endif
#ifdef EADDRNOTAVAIL
    { EADDRNOTAVAIL, "Local address not available" },
#endif
#ifdef EAFNOSUPPORT
    { EAFNOSUPPORT, "Address family not supported" },
#endif
#ifdef EAGAIN
    { EAGAIN, "Resource temporarily unavailable" },
#endif
#ifdef EBADF
    { EBADF, "Bad file descriptor" },
#endif
#ifdef EBUSY
    { EBUSY, "Resource busy" },
#endif
#ifdef ECANCELED
    { ECANCELED, "Operation canceled" },
#endif
#ifdef ECHILD
    { ECHILD, "No child processes" }, /* FIX?? */
#endif
#ifdef ECONNABORTED
    { ECONNABORTED, "Connection aborted" },
#endif
#ifdef ECONNREFUSED
    { ECONNREFUSED, "Connection refused" },
#endif
#ifdef EDEADLK
    { EDEADLK, "Resource deadlock avoided" },
#endif
#ifdef EDOM
    { EDOM, "Domain error" },
#endif
#ifdef EEXIST
    { EEXIST, "File exists" },
#endif
#ifdef EFAULT
    { EFAULT, "Bad address" }, /* FIX?? */
#endif
#ifdef EFBIG
    { EFBIG, "File too large" },
#endif
#ifdef EINPROGRESS
    { EINPROGRESS, "Operation in progress" },
#endif
#ifdef EINTR
    { EINTR, "Interrupted function call" }, /* ?? */
#endif
#ifdef EINVAL
    { EINVAL, "Invalid argument" },
#endif
#ifdef EIO
    { EIO, "Input/output error" },
#endif
#ifdef EISDIR
    { EISDIR, "Is a directory" },
#endif
#ifdef EMFILE
    { EMFILE, "Too many open files" },
#endif
#ifdef EMLINK
    { EMLINK, "Too many links" },
#endif
#ifdef EMSGSIZE
    { EMSGSIZE, "Inappropriate message buffer length" }, /* ?? */
#endif
#ifdef ENAMETOOLONG
    { ENAMETOOLONG, "Filename too long" },
#endif
#ifdef ENETUNREACH
    { ENETUNREACH, "Network unreachable" },
#endif
#ifdef ENFILE
    { ENFILE, "Too many open files" },
#endif
#ifdef ENODEV
    { ENODEV, "No such device" },
#endif
#ifdef ENOENT
    { ENOENT, "No such file or directory" },
#endif
#ifdef ENOEXEC
    { ENOEXEC, "Exec format error" },
#endif
#ifdef ENOLCK
    { ENOLCK, "No locks available" },
#endif
#ifdef ENOMEM
    { ENOMEM, "Not enough space" },
#endif
#ifdef ENOSPC
    { ENOSPC, "No space left on device" },
#endif
#ifdef ENOSYS
    { ENOSYS, "Function not implemented" },
#endif
#ifdef ENOTDIR
    { ENOTDIR, "Not a directory" },
#endif
#ifdef ENOTEMPTY
    { ENOTEMPTY, "Directory not empty" },
#endif
#ifdef ENOTSOCK
    { ENOTSOCK, "Not a socket" },
#endif
#ifdef ENOTSUP
    { ENOTSUP, "Not supported" }, /* ?? */
#endif
#ifdef ENOTTY
    { ENOTTY, "Inappropriate I/O control operation" },
#endif
#ifdef ENXIO
    { ENXIO, "No such device or address" },
#endif
#ifdef EOPNOTSUPP
    { EOPNOTSUPP, "Operation not supported" },
#endif
#ifdef EPERM
    { EPERM, "Operation not permitted" },
#endif
#ifdef EPIPE
    { EPIPE, "Broken pipe" },
#endif
#ifdef EPROTONOSUPPORT
    { EPROTONOSUPPORT, "Protocol not supported" },
#endif
#ifdef ERANGE
    { ERANGE, "Result too large" },
#endif
#ifdef EROFS
    { EROFS, "Read-only file system" },
#endif
#ifdef ESPIPE
    { ESPIPE, "Invalid seek" },
#endif
#ifdef ESRCH
    { ESRCH, "No such process" },
#endif
#ifdef ETIMEDOUT
    { ETIMEDOUT, "Operation timed out" },
#endif
#ifdef EXDEV
    { EXDEV, "Improper link" },
#endif
    { 0, "No error" }
};


AValue ARaiseErrnoExceptionByNum(AThread *t, int classNum)
{
    int i;
    const char *msg;
    int errNum;

    msg = NULL;
    errNum = errno;
    
    for (i = 0; i < NUM_ERROR_MESSAGES; i++) {
        if (ErrorMessages[i].errNum == errNum) {
            msg = ErrorMessages[i].message;
            break;
        }
    }

    return ARaiseByNum(t, classNum, msg);
}
